class InvalidNamespaceError(Exception):
    error_name = "InvalidNamespaceError"


class LocalStorageValidationError(Exception):
    error_name = "LocalStorageValidationError"


class LocalStorageError(Exception):
    error_name = "LocalStorageError"
